#  AST.py --- Class hierarchy supporting abstract syntax tree for parsed
#             formulas
#
#  last modified: 01/13/04
#
#
#                     =====================================
#                                 Marc Provost
#                          (marc.provost@mail.mcgill.ca)
#
#                                  Credits to:
#                             Jean-Sebastien BOLDUC
#                              (jseb@cs.mcgill.ca)
#
#                          McGill University, Montreal
#                     =====================================
#
#******************************************************************************

# This module defines convenient names for all built-in Python types (see
# Python documentation):
import types
from math import pow
from copy import copy
from ASTVisitor import ASTVisitor

#******************************************************************************
# Utilities

#INPUT: <string>
#OUTPUT: <int>
#Order: O(length(string)) (in our case len(col) <= 3)
#Info: Function used to transform strings used
#      cell column references (e.g. A5) by integers
#      for efficiency reasons.  Note that 0 are removed
#      from possible integers to avoid ambiguity.
#      (i.e. if "A" is converted to 0, "AAA"=000=0,
#       0==0 while "A" != "AAA").
def stringToInt(col):
    base=ord("A")
    answer=0
    for l in xrange(len(col)):
        answer+=(ord(col[-(l+1)])-base+1)*int(pow(26,l))
    return answer

#INPUT: <string>
#OUTPUT: <int>
#Order: O(1) since integer assumed <= 26^3
#Info: Invert function of stringToInt()

def intToString(integer):
    ans=""
    base=ord("A")
    i=2
    for i in [2,1,0]:
        tmp=int(pow(26,i))
        tmpInteger=integer
        for j in xrange(i):
            tmpInteger-=int(pow(26,j))
        if tmp<=tmpInteger or i==0:
            charValue=(tmpInteger-tmpInteger%tmp)/tmp
            integer=integer-charValue*tmp
            char=chr(charValue+base-1)
            ans+=char
    return ans

#******************************************************************************
### class ASTNode

class ASTNode:
    """Base (abstract) class for all AST nodes such as Number,Functions,etc.
    """
    def __init__(self):
        #This class is abstract, and error is raised if an attempt is made
        #to instantiate it.
        if self.__class__ == ASTNode:
            raise NotImplementedError,"abstract class cannot be instantiated"

    #INPUT: <Visitor>
    #OUTPUT: <None>
    #Order: O(1)
    #Info: Function part of the Visitor pattern, depending on the
    #      class, call the appropriate method of ``visitor''.  Thus, it
    #      must be implemented at a lower level.
    def accept(self,visitor):
        pass

    #String representation
    def __str__(self):
        pass


#******************************************************************************
### class Number

class Number(ASTNode):
    """Encapsulate numerical data (float)
    """
    #Note: Instance of this class are leaf nodes of the AST

    #INPUT: <int>
    #Order: O(1)
    #Info: A TypeError exception is raised if the parameter
    #      number is not of the right type.
    def __init__(self, number):
        if type(number) not in (types.IntType, types.FloatType):
          raise TypeError,"constructor expects an integer or a float"

        self.__value = float(number)  # explicit casting into "float"

    #INPUT: <None>
    #OUTPUT: <int>
    #Order: O(1)
    #Info: return encapsulated value.
    def getValue(self):
        return self.__value

    #INPUT: <ASTVisitor>
    #OUTPUT: None
    #Order: O(1)
    #Info: Method used in the visitor pattern.
    #      Calls the appropriate action to perform while
    #      visiting this object.
    def accept(self,visitor):
        if not isinstance(visitor, ASTVisitor):
            raise TypeError, "accept method expects a Visitor"
        visitor.visitNumber(self)

    def __str__(self):
        return str(self.__value)

    def __repr__(self):
        return str(self)


#******************************************************************************
### class CellRef

class CellRef(ASTNode):
    """ Encapsulate a reference to a cell object
    """
    #Note: References can either be absolute or relative.

    #INPUT: <int> <int> <int> (used as bool) <int> (used as bool)
    #Order: O(1)
    #Info:  Given a positive column & row, instantiate
    #       the object.  Since the references can be
    #       relative to the cell holding the formula,
    #       two booleans are taken as inputs, which will
    #       be used to replace a relative reference
    #       by another when a formula is copied from cell to
    #       cell.  TypeErrors are raised if the types of the
    #       inputs are incorrect.  ValueErrors are raised
    #       if column or row is negative.

    def __init__(self, column, row, isAbsCol, isAbsRow):
        if type(column) != types.IntType:
            raise TypeError, "Constructor expect an integer for arg1 (column)"
        if type(row) != types.IntType:
            raise TypeError, "Constructor expect an integer for arg2 (row)"
        if type(isAbsCol) != types.IntType:
            raise TypeError, \
                  "Constructor expect an integer for arg3 (isAbsColumn)"
        if type(isAbsRow) != types.IntType:
            raise TypeError, "Constructor expect an integer for arg4 (isAbsRow)"

        if column<0:
            raise ValueError, "Invalid column value, must be greater than 0"
        if row<0:
            raise ValueError, "Invalid row value, must be greater than 0"

        self.__column=column
        self.__row=row
        self.__isAbsCol=isAbsCol
        self.__isAbsRow=isAbsRow

    #Next follows accessors methods to return private fields

    #INPUT: <None>
    #OUTPUT: <int>
    #Order: O(1)
    #Info: returns the row of this CellRef
    def getRow(self):
        return self.__row

    #INPUT: <None>
    #OUTPUT: <int>
    #Order: O(1)
    #Info: returns the column of this CellRef
    def getColumn(self):
        return self.__column

    #INPUT: <None>
    #OUTPUT: <int>
    #Order: O(1)
    #Info: returns true if the row is absolute
    def isAbsRow(self):
        return self.__isAbsRow

    #INPUT: <None>
    #OUTPUT: <int>
    #Order: O(1)
    #Info: returns true if the column is absolute
    def isAbsColumn(self):
        return self.__isAbsCol

    #INPUT: <ASTVisitor>
    #OUTPUT: None
    #Order: O(1)
    #Info: Method used in the visitor pattern.
    #      Calls the appropriate action to perform while
    #      visiting this object.
    def accept(self,visitor):
        if not isinstance(visitor, ASTVisitor):
            raise TypeError, "accept method expects a Visitor"
        visitor.visitCellRef(self)

    def __str__(self):
        absCol=""
        absRow=""
        if self.__isAbsCol:
            absCol="$"
        if self.__isAbsRow:
            absRow="$"
        return absCol+intToString(self.__column)+absRow+str(self.__row)

    def __repr__(self):
        return str(self)

class RangeRef(ASTNode):
    """Encapsulate a reference to a range of cells
    """

    def __init__(self, cornerA, cornerB):
        if not isinstance(cornerA, CellRef):
            raise TypeError,"constructor expects a CellRef for arg1"
        if not isinstance(cornerB, CellRef):
            raise TypeError,"constructor expects a CellRef for arg2"

        self.__cornerA=cornerA
        self.__cornerB=cornerB

    #INPUT: <None>
    #OUTPUT: <ListOfCellRef>
    #Order: O(rowA-rowB|*|colA-colB|)
    #Info: Returns a list of the CellRef expressed by
    #      the mulref, ordered in the direction cornerA->cornerB.
    #      For instance, A1:B7 return [A1,B1,...,A7,B7] while
    #      A7:B1 returns [A7,B7,...,A1,B1].
    def getCellRefSet(self):
        ans=[]
        incrCol=0
        incrRow=0

        colA=self.__cornerA.getColumn()
        colB=self.__cornerB.getColumn()
        rowA=self.__cornerA.getRow()
        rowB=self.__cornerB.getRow()

        if colA<=colB:
            incrCol=1
        if rowA<=rowB:
            incrRow=1

        if incrRow and incrCol:
            for i in xrange(rowA,rowB+1):
                for j in xrange(colA,colB+1):
                    ans.append(CellRef(j,i,0,0))
        elif incrRow and (not incrCol):
            for i in xrange(rowA,rowB+1):
                for j in xrange(colA,colB-1,-1):
                    ans.append(CellRef(j,i,0,0))
        elif (not incrRow) and incrCol:
            for i in xrange(rowA,rowB-1,-1):
                for j in xrange(colA,colB+1):
                    ans.append(CellRef(j,i,0,0))
        else:
            for i in xrange(rowA,rowB-1,-1):
                for j in xrange(colA,colB-1,-1):
                    ans.append(CellRef(j,i,0,0))
        return ans

    #INPUT: <None>
    #OUTPUT: <CellRef>
    #Order: O(1)
    #Info: returns the first CellRef of this RangeRef
    def getFirstCorner(self):
        return self.__cornerA

    #INPUT: <None>
    #OUTPUT: <CellRef>
    #Order: O(1)
    #Info: returns the second CellRef of this RangeRef
    def getSecondCorner(self):
        return self.__cornerB
    #*******************************************************

    #INPUT: <ASTVisitor>
    #OUTPUT: None
    #Order: O(1)
    #Info: Method used in the visitor pattern.
    #      Calls the appropriate action to perform while
    #      visiting this object.
    def accept(self, visitor):
        if not isinstance(visitor, ASTVisitor):
            raise TypeError, "accept method expects a Visitor"
        visitor.visitRangeRef(self)

    def __str__(self):
        return str(self.__cornerA)+":"+str(self.__cornerB)

#******************************************************************************
### class Function

class Function(ASTNode):
    """Encapsulate a function
    """

    #INPUT: <string>, <listOfASTNode>
    #Order: O(len(arguments))
    #Info: Initialize a Function object.  Note that the
    #      actual implementation of the Function is external to
    #      the AST.
    def __init__(self, name, arguments):

        if type(name) != types.StringType:
            raise TypeError,"constructor expects a str for the function name"
        if type(arguments) != types.ListType:
            raise TypeError,"constructor expects a list for the function args"

        for elem in arguments:
            if not isinstance(elem, ASTNode):
                raise TypeError,"invalid item in the arguments list"

        self.__arguments=copy(arguments)
        self.__name=name

    #INPUT: <None>
    #OUTPUT: <string>
    #Order: O(1)
    #Info: returns the name of this function.
    #      Note: A copy is not returned because strings
    #      are immutable in python
    def getName(self):
        return self.__name

    #INPUT: <None>
    #OUTPUT: <ListOfASTNode>
    #Order: O(1)
    #Info: returns the arguments of this function.  Note:
    #      A copy of the list is returned to prevent from
    #      directly modifying the datastructure.
    def getArgs(self):
        return copy(self.__arguments)

    #INPUT: <ASTVisitor>
    #OUTPUT: None
    #Order: O(1)
    #Info: Method used in the visitor pattern.
    #      Calls the appropriate action to perform while
    #      visiting this object.
    def accept(self,visitor):
        if not isinstance(visitor, ASTVisitor):
            raise TypeError, "accept method expects a Visitor"
        for argument in self.__arguments:
            argument.accept(visitor)
        visitor.visitFunction(self)

    def __str__(self):
        strRep=self.__name+"("
        for i in xrange(len(self.__arguments)):
            if i==0:
                strRep+=str(self.__arguments[i])
            else:
                strRep+=","+str(self.__arguments[i])
        strRep+=")"
        return strRep

#******************************************************************************








